
/* * * * * * * * * * * * * Author's note * * * * * * * * * * * *\
*   _       _   _       _   _       _   _       _     _ _ _ _   *
*  |_|     |_| |_|     |_| |_|_   _|_| |_|     |_|  _|_|_|_|_|  *
*  |_|_ _ _|_| |_|     |_| |_|_|_|_|_| |_|     |_| |_|_ _ _     *
*  |_|_|_|_|_| |_|     |_| |_| |_| |_| |_|     |_|   |_|_|_|_   *
*  |_|     |_| |_|_ _ _|_| |_|     |_| |_|_ _ _|_|  _ _ _ _|_|  *
*  |_|     |_|   |_|_|_|   |_|     |_|   |_|_|_|   |_|_|_|_|    *
*                                                               *
*                     http://www.humus.name                     *
*                                                                *
* This file is a part of the work done by Humus. You are free to   *
* use the code in any way you like, modified, unmodified or copied   *
* into your own work. However, I expect you to respect these points:  *
*  - If you use this file and its contents unmodified, or use a major *
*    part of this file, please credit the author and leave this note. *
*  - For use in anything commercial, please request my approval.     *
*  - Share your work and ideas too as much as you can.             *
*                                                                *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "App.h"

BaseApp *app = new App();

void App::resetCamera()
{
	camPos = vec3(-5.3f, 27.4f, 109.6f);
	wx = 0.19f;
	wy = 3.10f;
}

void App::moveCamera(const vec3 &dir)
{
	camPos += dir * (frameTime * speed);

	// Collision detection against terrain
	float xc = camPos.x + 127.5f;
	float yc = camPos.z + 127.5f;

	uint ix = (int) floorf(xc);
	uint iy = (int) floorf(yc);

	if (ix < 255 && iy < 255)
	{
		float fx = xc - ix;
		float fy = yc - iy;

		ubyte *pixels = m_TerrainImage.getPixels();

		float h00 = (float) pixels[(iy + 0) * 256 + (ix + 0)];
		float h01 = (float) pixels[(iy + 0) * 256 + (ix + 1)];
		float h10 = (float) pixels[(iy + 1) * 256 + (ix + 0)];
		float h11 = (float) pixels[(iy + 1) * 256 + (ix + 1)];

		float h0 = lerp(h00, h01, fx);
		float h1 = lerp(h10, h11, fx);
		float h = lerp(h0, h1, fy);

		h *= (50.0f / 255.0f);
		h += 1.0f;

		if (camPos.y < h)
			camPos.y = h;
	}
}

bool App::init()
{
	speed = 50;
	return true;
}

void App::exit()
{
}

void App::onSize(const int w, const int h)
{
	D3D10App::onSize(w, h);

	if (renderer)
	{
		// Make sure render targets are the size of the window
		renderer->resizeRenderTarget(m_DepthRT,     w, h, 1, 1, 1);
		renderer->resizeRenderTarget(m_DepthCopyRT, w, h, 1, 1, 1);
	}
}

bool App::onMouseButton(const int x, const int y, const MouseButton button, const bool pressed)
{
	if (BaseApp::onMouseButton(x, y, button, pressed))
		return true;

	m_MousePressed = pressed;

	if (pressed)
	{
		if (button == MOUSE_LEFT)
		{
			Road road(camPos, camPos);
			m_Roads.add(road);
		}
		else if (button == MOUSE_RIGHT)
		{
			m_Roads.clear();
		}
	}

	m_HelpDisplayTime = 0.0f;

	return true;
}

bool App::initAPI()
{
	// Override the user's MSAA settings
	return D3D10App::initAPI(DXGI_FORMAT_R8G8B8A8_UNORM, DXGI_FORMAT_UNKNOWN, 1, NO_SETTING_CHANGE);
}

void App::exitAPI()
{
	D3D10App::exitAPI();
}

bool App::load()
{
	// Shaders
	if ((m_TerrainShader = renderer->addShader("terrain.shd")) == SHADER_NONE) return false;
	if ((m_SkyboxShader = renderer->addShader("skybox.shd")) == SHADER_NONE) return false;
	if ((m_RoadShader = renderer->addShader("road.shd")) == SHADER_NONE) return false;
	if ((m_RoadStencilShader = renderer->addShader("roadstencil.shd")) == SHADER_NONE) return false;
	if ((m_ResolveDepthShader = renderer->addShader("resolvedepth.shd")) == SHADER_NONE) return false;

	// Depth buffer and copy
	if ((m_DepthRT = renderer->addRenderDepth(width, height, 1, FORMAT_D24S8, msaaSamples, SS_NONE, SAMPLE_DEPTH)) == TEXTURE_NONE) return false;
	if ((m_DepthCopyRT = renderer->addRenderTarget(width, height, 1, 1, 1, FORMAT_R32F, 1, SS_NONE)) == TEXTURE_NONE) return false;

	// Samplerstates
	if ((m_TrilinearClamp = renderer->addSamplerState(TRILINEAR, CLAMP, CLAMP, CLAMP)) == SS_NONE) return false;
	if ((m_TrilinearAniso = renderer->addSamplerState(TRILINEAR_ANISO, WRAP, WRAP, WRAP)) == SS_NONE) return false;

	// Terrain
	m_TerrainImage.loadDDS("Terrain.dds");
	if ((m_Terrain = renderer->addTexture(m_TerrainImage, false, linearClamp)) == TEXTURE_NONE) return false;
	m_TerrainImage.uncompressImage();

	if ((m_TerrainShadow = renderer->addTexture("TerrainShadow.dds", false, linearClamp)) == TEXTURE_NONE) return false;

	// Textures
	if ((m_Ground0 = renderer->addTexture("../Textures/Grass2.jpg", true, m_TrilinearAniso)) == TEXTURE_NONE) return false;
	if ((m_Ground1 = renderer->addTexture("../Textures/rock.pcx", true, m_TrilinearAniso)) == TEXTURE_NONE) return false;
	if ((m_Road = renderer->addTexture("../Textures/road_free.dds", true, m_TrilinearAniso)) == TEXTURE_NONE) return false;



	const char *files[] =
	{
		"../Textures/CubeMaps/Mountain/posx.jpg", "../Textures/CubeMaps/Mountain/negx.jpg",
		"../Textures/CubeMaps/Mountain/posy.jpg", "../Textures/CubeMaps/Mountain/negy.jpg",
		"../Textures/CubeMaps/Mountain/posz.jpg", "../Textures/CubeMaps/Mountain/negz.jpg",
	};
	if ((m_Environment = renderer->addCubemap(files, true, m_TrilinearClamp)) == TEXTURE_NONE) return false;

	// Blendstates
	if ((m_BlendMulAlpha = renderer->addBlendState(DST_ALPHA, ONE_MINUS_SRC_ALPHA)) == BS_NONE) return false;

	// Rasterizerstates
	if ((m_NoMSAA = renderer->addRasterizerState(CULL_NONE, SOLID, false, false)) == RS_NONE) return false;

	// Depthstates
	if ((m_StencilTag  = renderer->addDepthState(true,  false, LEQUAL, true, 0xFF, 0xFF, ALWAYS, ALWAYS, KEEP, KEEP, INCR, DECR, KEEP, KEEP)) == DS_NONE) return false;
	if ((m_StencilCull = renderer->addDepthState(false, false, LEQUAL, true, 0xFF, 0xFF, NOTEQUAL, NOTEQUAL, KEEP, KEEP, KEEP, KEEP, SET_ZERO, SET_ZERO)) == DS_NONE) return false;

	// Index buffer
	ushort indices[] =
	{
		0, 2, 3,   0, 3, 1,
		1, 3, 7,   1, 7, 5,
		5, 7, 6,   5, 6, 4,
		4, 6, 2,   4, 2, 0,
		2, 6, 7,   2, 7, 3,
		4, 0, 1,   4, 1, 5,
	};
	if ((m_BoxIB = renderer->addIndexBuffer(elementsOf(indices), 2, STATIC, indices)) == IB_NONE) return false;

	// Add an initial road
	Road road(float3(-11.7f, 28.8f, -127.6f), float3(-5.08f, 28.7f,  127.6f));
	m_Roads.add(road);

	m_MousePressed = false;

	m_HelpDisplayTime = 5.0f;

	return true;
}

void App::unload()
{

}

void App::drawFrame()
{
	float4x4 projection = toD3DProjection(perspectiveMatrixX(1.5f, width, height, 0.1f, 500));
	float4x4 modelview = rotateXY(-wx, -wy);
	float4x4 invMvpEnv = !(projection * modelview);
	modelview.translate(-camPos);
	float4x4 mvp = projection * modelview;


	renderer->changeRenderTarget(FB_COLOR, m_DepthRT);
	renderer->clear(false, true, true);

	/*
		Terrain
	*/
	renderer->reset();
	renderer->setRasterizerState(cullBack);
	renderer->setShader(m_TerrainShader);
	renderer->setShaderConstant4x4f("mvp", mvp);
	renderer->setTexture("terrain", m_Terrain);
	renderer->setTexture("ground0", m_Ground0);
	renderer->setTexture("ground1", m_Ground1);
	renderer->setTexture("shadow", m_TerrainShadow);
	renderer->setSamplerState("filter", m_TrilinearAniso);
	renderer->setSamplerState("shadow_filter", linearClamp);
	renderer->setSamplerState("vsfilter", linearClamp);
	renderer->apply();

	device->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP);
	device->DrawInstanced(256, 255, 0, 0);


	/*
		Skybox
	*/
	renderer->reset();
	renderer->setRasterizerState(m_NoMSAA);
	renderer->setShader(m_SkyboxShader);
	renderer->setShaderConstant4x4f("invMvp", invMvpEnv);
	renderer->setTexture("env", m_Environment);
	renderer->setSamplerState("filter", m_TrilinearClamp);
	renderer->setDepthState(noDepthWrite);
	renderer->apply();

	device->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
	device->Draw(3, 0);


	/*
		Resolve Depth
	*/
	renderer->changeRenderTarget(m_DepthCopyRT);

		renderer->reset();
		renderer->setRasterizerState(m_NoMSAA);
		renderer->setShader(m_ResolveDepthShader);
		renderer->setTexture("Tex", m_DepthRT);
		renderer->setDepthState(noDepthTest);
		renderer->apply();

		device->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
		device->Draw(3, 0);

	renderer->changeRenderTarget(FB_COLOR, m_DepthRT);


	uint count = m_Roads.getCount();

	// Update road being placed
	if (m_MousePressed && count > 0)
		m_Roads[count - 1].m_Pos1 = camPos;

	for (uint i = 0; i < count; i++)
	{
		/*
			Compute transformation from screen space to the box between two selected points
			such that the resulting .xz coordinates contain a texture coordinate
		*/
		float3 v0 = m_Roads[i].m_Pos0;
		float3 v1 = m_Roads[i].m_Pos1;

		float len = distance(v0, v1);
		float3 mid = 0.5f * (v0 + v1);

		float3 dir_x = normalize(v1 - v0);
		float3 dir_z = normalize(cross(dir_x, float3(0, 1, 0)));
		float3 dir_y = normalize(cross(dir_z, dir_x));

		const float size = 30.0f; // Relatively large number, ideally should be computed to be only as big as necessary for a road segment
		const float down_factor = 0.3f; // Move the box down a little to better use the box volume in normal cases for this demo

		float4x4 rot(
			dir_x.x, dir_y.x, dir_z.x, mid.x,
			dir_x.y, dir_y.y, dir_z.y, mid.y - down_factor * size,
			dir_x.z, dir_y.z, dir_z.z, mid.z,
			0, 0, 0, 1
		);

		float4x4 trans = rot * scale(len * 0.5f, size, 2);
		float4x4 transform = mvp * trans;

		float4x4 reverse_transform = translate(0.5f, 0.0f, 0.5f) * scale(0.125f * len, 0.0f, -0.5f) * (!transform) * (translate(-1.0f, 1.0f, 0.0f) * scale(2.0f / width, -2.0f / height, 1.0f));


		/*
			Road stencil pass
		*/
		renderer->reset();
		renderer->setRasterizerState(cullNone);
		renderer->setDepthState(m_StencilTag);
		renderer->setShader(m_RoadStencilShader);
		renderer->setShaderConstant4x4f("transform", transform);
		renderer->setIndexBuffer(m_BoxIB);
		renderer->apply();

		device->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
		device->DrawIndexed(36, 0, 0);

		/*
			Road shader pass
		*/
		renderer->reset();
		renderer->setRasterizerState(cullFront);
		renderer->setDepthState(m_StencilCull);
		renderer->setBlendState(m_BlendMulAlpha);
		renderer->setShader(m_RoadShader);
		renderer->setTexture("Depth", m_DepthCopyRT);
		renderer->setTexture("Road", m_Road);
		renderer->setSamplerState("filter", m_TrilinearAniso);
		renderer->setShaderConstant4x4f("transform", transform);
		renderer->setShaderConstant4x4f("reverse_transform", reverse_transform);
		renderer->setIndexBuffer(m_BoxIB);
		renderer->apply();

		device->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
		device->DrawIndexed(36, 0, 0);
	}

	// Display help text
	if (m_HelpDisplayTime > 0)
	{
		if (configDialog->isVisible())
		{
			m_HelpDisplayTime = 0;
		}
		else
		{
			m_HelpDisplayTime -= min(frameTime, 0.1f);

			renderer->drawText("Press and hold left mouse button to place a new road.\nPress right mouse button to clear all roads.",
				width * 0.5f - 350, height * 0.5f - 40, 30, 38, defaultFont, linearClamp, blendSrcAlpha, noDepthTest);
		}
	}
}
